import attr
import structlog

from elgas import application, exceptions

LOG = structlog.get_logger("state")


class _SentinelBase(type):
    """
    Sentinel values

     - Inherit identity-based comparison and hashing from object
     - Have a nice repr
     - Have a *bonus property*: type(sentinel) is sentinel

     The bonus property is useful if you want to take the return value from
     next_event() and do some sort of dispatch based on type(event).

     Taken from h11.
    """

    def __repr__(self):
        return self.__name__


def make_sentinel(name):
    cls = _SentinelBase(name, (_SentinelBase,), {})
    cls.__class__ = cls
    return cls


IDLE = make_sentinel("IDLE")
AWAITING_READ_ACTUAL_VALUES_RESPONSE = make_sentinel(
    "AWAITING_READ_ACTUAL_VALUES_RESPONSE"
)
AWAITING_READ_DEVICE_PARAMETERS_RESPONSE = make_sentinel(
    "AWAITING_READ_DEVICE_PARAMETERS_RESPONSE"
)
AWAITING_READ_TIME_RESPONSE = make_sentinel("AWAITING_READ_TIME_RESPONSE")
AWAITING_READ_ARCHIVE_BY_TIME_RESPONSE = make_sentinel(
    "AWAITING_READ_ARCHIVE_BY_TIME_RESPONSE"
)
AWAITING_READ_ARCHIVE_RESPONSE = make_sentinel("AWAITING_READ_ARCHIVE_RESPONSE")

AWAITING_WRITE_TIME_RESPONSE = make_sentinel("AWAITING_WRITE_TIME_RESPONSE")
NEED_DATA = make_sentinel("NEED_DATA")


ELGAS_STATE_TRANSITIONS = {
    IDLE: {
        application.ReadInstantaneousValuesRequest: AWAITING_READ_ACTUAL_VALUES_RESPONSE,
        application.ReadDeviceParametersRequest: AWAITING_READ_DEVICE_PARAMETERS_RESPONSE,
        application.ReadTimeRequest: AWAITING_READ_TIME_RESPONSE,
        application.ReadArchiveByTimeRequest: AWAITING_READ_ARCHIVE_BY_TIME_RESPONSE,
        application.ReadArchiveRequest: AWAITING_READ_ARCHIVE_RESPONSE,
        application.WriteTimeRequest: AWAITING_WRITE_TIME_RESPONSE,
    },
    AWAITING_READ_ACTUAL_VALUES_RESPONSE: {
        application.ReadInstantaneousValuesResponse: IDLE,
    },
    AWAITING_READ_DEVICE_PARAMETERS_RESPONSE: {
        application.ReadDeviceParametersResponse: IDLE
    },
    AWAITING_READ_TIME_RESPONSE: {application.ReadTimeResponse: IDLE},
    AWAITING_READ_ARCHIVE_BY_TIME_RESPONSE: {
        application.ReadArchiveByTimeResponse: IDLE
    },
    AWAITING_READ_ARCHIVE_RESPONSE: {application.ReadArchiveResponse: IDLE},
    AWAITING_WRITE_TIME_RESPONSE: {application.WriteTimeResponse: IDLE},
}


@attr.s(auto_attribs=True)
class ElgasState:
    """
    Handles state changes in ELGAS Protocol
    """

    current_state: _SentinelBase = attr.ib(default=IDLE)

    def process_event(self, event):

        event_type = type(event)

        try:
            new_state = ELGAS_STATE_TRANSITIONS[self.current_state][event_type]
        except KeyError:
            raise exceptions.LocalElgasProtocolError(
                f"Can't handle {event_type} when state={self.current_state}"
            )
        old_state = self.current_state
        self.current_state = new_state
        LOG.debug(f"Elgas state transitioned from {old_state} to {new_state}")



from enum import IntEnum


class DeviceType(IntEnum):
    ELCOR = 128
    ELCOR_PLUS = 131
    ELCOR_LITE = 134


class CertificationVariant(IntEnum):
    CMI_MID = 2
    GENERIC = 4


class SwitchFunction(IntEnum):
    NONE = 0
    PROTECT_METROLOGICAL = 1
    PROTECT_ALL = 2


class CompressibilityFormula(IntEnum):
    CONST = 0
    AGANX19 = 1
    SGERG88 = 2
    AGANX19MOD = 3
    AGA8G1 = 4
    AGA8G2 = 5
    AGA892DC = 6
    GOSTNX19MOD = 7


class ParameterObjectType(IntEnum):

    SYSTEM_PARAMETER = 0
    ANALOG_MEASURAND = 30
    BINARY = 31
    COUNTER = 32
    STANDARD_COUNTER = 33
    FLOW_RATE = 34
    STANDARD_FLOW_RATE = 35
    CONVERSION_COEFFICIENT = 36
    ERROR_COUNTER = 45
    ERROR_STANDARD_COUNTER = 46
    COMPRESSIBILITY = 47
    TIME_WINDOW = 48
    CORRECTION_COUNTER = 49
    DOUBLE_COUNTER = 53
    DOUBLE_ERROR_COUNTER = 54
    DIAGNOSTICS = 59
    DEVICE_ERROR = 61
    SUM_OF_ALARMS = 62
    TIMER = 65

    TARIFF_COUNTER = 68
    BASE_TARIFF_COUNTER = 69
    SET_POINT = 70
    DIFFERENCE_COUNTER = 72
    DIFFERENCE_BASE_COUNTER = 73
    COMPRESSIBILITY_Z = 74
    COMPRESSIBILITY_Z_BASE = 75
    ENERGY = 77
    ERROR_ENERGY = 78
    DOUBLE_TARIFF_COUNTER = 79
    ANALOG_STATISTICS = 80
    COUNTER_STATISTICS = 81
    STANDARD_COUNTER_STATISTICS = 82
    STATISTICS = 83
    ANALOG_TIME_STATISTICS = 84
    TIME_STATISTICS = 85

    MODEM = 141



import random
import sys
from datetime import datetime
from typing import *


class SecretStr(str):
    """
    String that censors its __repr__ if called from another repr.
    """

    def __repr__(self):
        """Produce a string representation."""
        frame_depth = 1

        try:
            while True:
                f = sys._getframe(frame_depth)
                frame_depth += 1

                if f.f_code.co_name == "__repr__":
                    return "<SECRET>"
        except ValueError:
            pass

        return super().__repr__()


class SecretByte(bytes):
    """
    Bytes that censors its __repr__ if called from another repr.
    """

    def __repr__(self):
        """Produce a string representation."""
        frame_depth = 1

        try:
            while True:
                f = sys._getframe(frame_depth)
                frame_depth += 1

                if f.f_code.co_name == "__repr__":
                    return "<SECRET>"
        except ValueError:
            pass

        return super().__repr__()


def to_secret_byte(data: bytes) -> SecretByte:
    if data:
        return SecretByte(data)


def to_secret_str(data: str) -> SecretStr:
    return SecretStr(data)


def calculate_lrc(data: bytearray) -> int:
    lrc = 0
    for _byte in data:
        temp = lrc ^ _byte
        lrc = temp & 0xFF
    return lrc & 0xFF


def calculate_checksum(data: bytearray) -> int:
    checksum = 0
    for _byte in data:
        temp = checksum + _byte
        checksum = temp & 0xFF
    return checksum & 0xFF


def calculate_drc(data: bytearray) -> int:
    drc = 0
    for _byte in data:
        temp = drc << 1
        if temp & 0x100:
            temp = temp + 1
        temp = temp & 0xFF
        temp = temp ^ _byte
        drc = temp & 0xFF
    return drc & 0xFF


def generate_crc16_table():
    result = []
    for byte in range(256):
        crc = 0x0000
        for _ in range(8):
            if (byte ^ crc) & 0x0001:
                crc = (crc >> 1) ^ 0xA001
            else:
                crc >>= 1
            byte >>= 1
        result.append(crc)
    return result


crc16_table = generate_crc16_table()


def calculate_crc(data: bytearray):
    crc = 0xFFFF
    for a in data:
        idx = crc16_table[(crc ^ a) & 0xFF]
        crc = ((crc >> 8) & 0xFF) ^ idx
    result = ((crc << 8) & 0xFF00) | ((crc >> 8) & 0x00FF)
    return result


def escape_characters(data: bytes) -> bytes:
    """
    Characters that are used for telegram control are replaced with others
    so that it easier to parse the message over the wire.
    Final character is never replaced.

    '\x0d' -> '\x1b\x0e'
    '\x1b' -> '\x1b\x1b'
    '\x8d' -> '\x1b\x0f'
    """
    if not data.endswith(b"\x0d"):
        raise ValueError("Data does not end with end-char 0x0D")
    workable_data = data[:-1]
    escaped_data = (
        workable_data.replace(b"\x1b", b"\x1b\x1b")
        .replace(b"\x0d", b"\x1b\x0e")
        .replace(b"\x8d", b"\x1b\x0f")
    )
    return escaped_data + b"\x0d"


def return_characters(data: bytes) -> bytes:
    """
    Characters that are used for telegram control are replaced with others
    so that it is easier to parse the message over the wire.
    Final character is never replaced.
    This function returns them to their original form.

    '\x1b\x0e' -> '\x0d'
    '\x1b\x1b' -> '\x1b'
    '\x1b\x0f' -'\x8d'

    It is important to only do one pass as the first pass can result in new combinations
    of escape sequence. Ex '\x1b\x1b\x0e' -> '\x1b\x0e'
    """
    in_data = bytearray(data)
    out = bytearray()
    while in_data:
        byte = in_data.pop(0)

        if byte == 0x1B:
            next_byte = in_data.pop(0)
            if next_byte == 0x0E:
                out.append(0x0D)
            elif next_byte == 0x1B:
                out.append(0x1B)
            elif next_byte == 0x0F:
                out.append(0x8D)
            else:
                out.append(byte)
                out.append(next_byte)
        else:
            out.append(byte)

    return bytes(out)


def pad_password(password: str):
    """
    Passwords are 6 characters long. But are sent as 10 characters where the last
    4 characters are insignificant.
    """
    possible_chars = "abcdefghijklmnopqrstuvxyzABCDEFGHIJKLMNOPQRSTUVXYZ0123456789"
    two_random = random.choices(possible_chars, k=2)
    # return password + "".join(four_random)
    return password + "!\x03" + "".join(two_random)


def bytes_to_datetime(data: bytes) -> Tuple[datetime, bool, bool]:
    """
    BCD of 6 bytes  seconds, minutes, hour, day, month, year
    10 33 12 30 05 06 == 2006-05-30T12:33:10
    """
    if len(data) != 6:
        raise ValueError(f"{data!r} is not a BCD encoded datetime")
    second = from_bcd(data[0:1])
    minute = from_bcd(data[1:2])
    hour_data = data[2]
    hour_byte = (hour_data & 0b00111111).to_bytes(1, "little")
    hour = from_bcd(hour_byte)
    is_dst = bool(hour_data & 0b10000000)
    supports_dst = bool(hour_data & 0b01000000)
    day = from_bcd(data[3:4])
    month = from_bcd(data[4:5])
    # The year is only the last 2 digits. We set it so all dates are in the 20xx
    # I will probably be retired when it becomes a problem :)
    year = from_bcd(data[5:]) + 2000

    return (
        datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        ),
        is_dst,
        supports_dst,
    )


def datetime_to_bytes(timestamp: datetime) -> bytes:
    """
    Packs a datetime into a BCD encoded format
    """
    out = bytearray()
    out.extend(to_bdc(timestamp.second))
    out.extend(to_bdc(timestamp.minute))
    out.extend(to_bdc(timestamp.hour))
    out.extend(to_bdc(timestamp.day))
    out.extend(to_bdc(timestamp.month))
    out.extend(to_bdc(timestamp.year - 2000))

    return bytes(out)


def to_bdc(number: int) -> bytes:
    """
    4 bit bcd (Binary Coded Decimal)
    Example: Decimal 30 would be encoded with b"\x30" or 0b0011 0000
    """
    chars = str(number)
    if (len(chars) % 2) != 0:
        # pad string to make it a hexadecimal one.
        chars = "0" + chars
    bcd = bytes.fromhex(str(chars))
    return bcd


def from_bcd(data: bytes) -> int:
    """
    make a bcd encoded bytestring into an integer
    Example: b"\x30" should be 30
    """
    chars = data.hex()
    return int(chars)


def pop_many(array: bytearray, amount: int) -> bytearray:
    if amount < 2:
        raise ValueError("Only use pop_many if you need to get more than 1 byte")
    out = bytearray()
    for _ in range(0, amount):
        out.append(array.pop(0))
    return out


def pretty_text(in_data: bytes) -> str:
    """
    To removed data that should not be present in the text but might be on bit-level
    """
    termination_char = b"\x00"

    try:
        last_termination_char = in_data.rindex(termination_char)
        in_data = in_data[:last_termination_char]
    except ValueError:
        # no termination char in text
        pass
    return in_data.replace(b"\x00", b" ").lstrip().rstrip().decode("latin-1")


def parse_ip_address(data: bytes) -> str:
    if len(data) != 4:
        raise ValueError(f"IP addreess data should be 4 bytes, got {len(data)}")
    numbers = [str(x) for x in data]
    return ".".join(numbers)



from typing import *

import attr

import elgas.parameters
import elgas.parameters.enumerations
import elgas.parameters.factory
from elgas.utils import pop_many


@attr.s(auto_attribs=True)
class SimpleObjectHolder:
    """Just a class to hold data to verify in development."""

    length: int
    object_type: int
    data: bytes


@attr.s(auto_attribs=True)
class ScadaParameterParser:
    data: Optional[bytes] = attr.ib(default=None)

    def parse(self, data: Optional[bytes] = None):
        working_data = bytearray(self.data or data)

        if not working_data:
            raise ValueError("No data to parse")

        object_list = list()

        while working_data:
            length = int.from_bytes(pop_many(working_data, 2), "little")
            object_type = elgas.parameters.enumerations.ParameterObjectType(
                working_data.pop(0)
            )
            data = working_data[: length - 3]
            object_list.append(
                elgas.parameters.factory.ParameterFactory.from_bytes(object_type, data)
            )

            working_data = working_data[length - 3 :]

        return object_list



import struct
from typing import ClassVar

import attr
import marshmallow
from marshmallow import post_load

from elgas.parameters.enumerations import (
    CertificationVariant,
    CompressibilityFormula,
    DeviceType,
    ParameterObjectType,
    SwitchFunction,
)
from elgas.parameters.gas_composition import GasComposition, GasCompositionSchema
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class SystemParameters:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.SYSTEM_PARAMETER

    device_type: int
    serial_number: int
    firmware_version: str  # TODO: Length 5
    service_version: int  # Used in how to parse the object.
    certification_variant: int
    station_id: str  # TODO: Lenght 17
    password_for_full_access_active: bool
    password_for_reading_is_on: bool
    metrological_switch: bool
    user_switch: bool
    switch_function: int
    parameter_crc: int  # to check if parameters have been changed from a previous read.
    measuring_period: int  # seconds
    archive_period: int  # seconds
    base_pressure: float
    base_temperature: float
    compressibility_formula: int
    gas_composition: GasComposition
    data_archive_record_length: int
    binary_archive_record_length: int
    daily_archive_record_length: int
    monthly_archive_record_length: int
    instantaneous_values_error_bit_order: int
    binary_archive_record_error_bit_order: int
    data_archive_record_error_bit_order: int
    optical_port_speed: int
    optical_port_protocol: int
    optical_port_bit_control: int
    communication_port_0_speed: int
    communication_port_0_protocol: int
    communication_port_0_bit_control: int
    communication_port_1_speed: int
    communication_port_1_protocol: int
    communication_port_1_bit_control: int
    communication_port_2_speed: int
    communication_port_2_protocol: int
    communication_port_2_bit_control: int
    gas_day_hour: int
    fixed_barometric_pressure: float
    altitude: float
    status_archive_record_length: int
    pressure_unit_type: int  # TODO: how to interpret?
    pressure_unit_text: str
    temperature_unit_type: int  # TODO: how to interpret?
    temperature_unit_text: str
    altitude_unit_type: int
    altitude_unit_text: str
    gross_calorific_value_type: int
    gross_calorific_value_text: str
    dst_region: int
    gmt_hour_shift: int
    billing_archive_record_length: int
    conditions_for_combustion_heat: int
    device_variant: int
    bit_control: int
    places_for_corrected_volume_counters: int
    device_features_bits: int
    device_features: str
    version_metrological_part: str
    metrological_crc: int
    metrological_crc_32: int
    application_crc_32: int

    @classmethod
    def from_bytes(cls, in_bytes: bytes):

        # first 2 bytes is length and not included in parsing
        # third byte is object type and not included in parsing.
        data = bytearray(in_bytes)
        device_type = data.pop(0)
        serial_number = int.from_bytes(pop_many(data, 4), "little")
        firmware_version = pretty_text(pop_many(data, 5))
        service_version = data.pop(0)
        certification_variant = data.pop(0)
        station_id = pretty_text(pop_many(data, 17))
        data_access = data.pop(0)
        password_for_full_access_active = bool(data_access & 0b00000001)
        password_for_reading_is_on = bool(data_access & 0b00000010)
        metrological_switch = bool(data_access & 0b00000100)
        user_switch = bool(data_access & 0b00001000)
        switch_function = int((data_access & 0b00110000) >> 4)
        parameter_crc = int.from_bytes(pop_many(data, 2), "little")
        measuring_period = data.pop(0)
        archiving_period = int.from_bytes(pop_many(data, 2), "little")
        base_pressure = struct.unpack("<f", pop_many(data, 4))[0]
        base_temperature = struct.unpack("<f", pop_many(data, 4))[0]
        compressibility_formula = data.pop(0)
        gas_composition = GasComposition.from_bytes(pop_many(data, (23 * 4)))
        # The record lengths are including the header with time.
        data_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        binary_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        daily_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        monthly_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        instantaneous_values_error_bit_order = int.from_bytes(
            pop_many(data, 2), "little"
        )
        binary_archive_record_error_bit_order = int.from_bytes(
            pop_many(data, 2), "little"
        )
        data_archive_record_error_bit_order = int.from_bytes(
            pop_many(data, 2), "little"
        )
        optical_port_speed = data.pop(0)
        optical_port_protocol = data.pop(0)
        optical_port_bit_control = data.pop(0)
        communication_port_0_speed = data.pop(0)
        communication_port_0_protocol = data.pop(0)
        communication_port_0_bit_control = data.pop(0)
        communication_port_1_speed = data.pop(0)
        communication_port_1_protocol = data.pop(0)
        communication_port_1_bit_control = data.pop(0)
        communication_port_2_speed = data.pop(0)
        communication_port_2_protocol = data.pop(0)
        communication_port_2_bit_control = data.pop(0)
        gas_hour = data.pop(0)
        fixed_barometric_pressure = struct.unpack("<f", pop_many(data, 4))[0]
        altitude = struct.unpack("<f", pop_many(data, 4))[0]
        _ = pop_many(data, 17)  # Not used
        status_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        pressure_unit_type = data.pop(0)
        pressure_unit_text = pretty_text(pop_many(data, 8))
        temperature_unit_type = data.pop(0)
        temperature_unit_text = pretty_text(pop_many(data, 8))
        altitude_unit_type = data.pop(0)
        altitude_unit_text = pretty_text(pop_many(data, 8))
        gross_calorific_value_type = data.pop(0)
        gross_calorific_value_text = pretty_text(pop_many(data, 8))
        dst_region = data.pop(0)  # seems like it should not be used.
        gmt_hour_shift = data.pop(0)
        billing_archive_record_length = int.from_bytes(pop_many(data, 2), "little")
        conditions_for_combustion_heat = data.pop(0)
        device_variant = data.pop(0)  # czech translation is device_value?
        _ = data.pop(0)  # not used
        bit_control = data.pop(0)
        corrected_volume_counters_amount = data.pop(0)
        device_features_bits = data.pop(0)
        device_features = pop_many(data, 16).hex()
        version_metrological_part = pretty_text(pop_many(data, 5))
        metrological_crc = int.from_bytes(pop_many(data, 2), "little")
        metrological_crc_32 = int.from_bytes(pop_many(data, 4), "little")
        application_crc_32 = int.from_bytes(pop_many(data, 4), "little")

        assert not data

        return cls(
            device_type=device_type,
            serial_number=serial_number,
            firmware_version=firmware_version,
            service_version=service_version,
            certification_variant=certification_variant,
            station_id=station_id,
            password_for_full_access_active=password_for_full_access_active,
            password_for_reading_is_on=password_for_reading_is_on,
            metrological_switch=metrological_switch,
            user_switch=user_switch,
            switch_function=switch_function,
            parameter_crc=parameter_crc,
            measuring_period=measuring_period,
            archive_period=archiving_period,
            base_pressure=base_pressure,
            base_temperature=base_temperature,
            compressibility_formula=compressibility_formula,
            gas_composition=gas_composition,
            data_archive_record_length=data_archive_record_length,
            binary_archive_record_length=binary_archive_record_length,
            daily_archive_record_length=daily_archive_record_length,
            monthly_archive_record_length=monthly_archive_record_length,
            instantaneous_values_error_bit_order=instantaneous_values_error_bit_order,
            binary_archive_record_error_bit_order=binary_archive_record_error_bit_order,
            data_archive_record_error_bit_order=data_archive_record_error_bit_order,
            optical_port_speed=optical_port_speed,
            optical_port_protocol=optical_port_protocol,
            optical_port_bit_control=optical_port_bit_control,
            communication_port_0_speed=communication_port_0_speed,
            communication_port_0_protocol=communication_port_0_protocol,
            communication_port_0_bit_control=communication_port_0_bit_control,
            communication_port_1_speed=communication_port_1_speed,
            communication_port_1_protocol=communication_port_1_protocol,
            communication_port_1_bit_control=communication_port_1_bit_control,
            communication_port_2_speed=communication_port_2_speed,
            communication_port_2_protocol=communication_port_2_protocol,
            communication_port_2_bit_control=communication_port_2_bit_control,
            gas_day_hour=gas_hour,
            fixed_barometric_pressure=fixed_barometric_pressure,
            altitude=altitude,
            status_archive_record_length=status_archive_record_length,
            pressure_unit_type=pressure_unit_type,
            pressure_unit_text=pressure_unit_text,
            temperature_unit_type=temperature_unit_type,
            temperature_unit_text=temperature_unit_text,
            altitude_unit_type=altitude_unit_type,
            altitude_unit_text=altitude_unit_text,
            gross_calorific_value_type=gross_calorific_value_type,
            gross_calorific_value_text=gross_calorific_value_text,
            dst_region=dst_region,
            gmt_hour_shift=gmt_hour_shift,
            billing_archive_record_length=billing_archive_record_length,
            conditions_for_combustion_heat=conditions_for_combustion_heat,
            device_variant=device_variant,
            bit_control=bit_control,
            places_for_corrected_volume_counters=corrected_volume_counters_amount,
            device_features_bits=device_features_bits,
            device_features=device_features,
            version_metrological_part=version_metrological_part,
            metrological_crc=metrological_crc,
            metrological_crc_32=metrological_crc_32,
            application_crc_32=application_crc_32,
        )


class SystemParametersSchema(marshmallow.Schema):

    device_type = marshmallow.fields.Integer(required=True)
    serial_number = marshmallow.fields.Integer(required=True)
    firmware_version = marshmallow.fields.String(required=True)
    service_version = marshmallow.fields.Integer(required=True)
    certification_variant = marshmallow.fields.Integer(required=True)
    station_id = marshmallow.fields.String(required=True)
    password_for_full_access_active = marshmallow.fields.Boolean(required=True)
    password_for_reading_is_on = marshmallow.fields.Boolean(required=True)
    metrological_switch = marshmallow.fields.Boolean(required=True)
    user_switch = marshmallow.fields.Boolean(required=True)
    switch_function = marshmallow.fields.Integer(required=True)
    parameter_crc = marshmallow.fields.Integer(required=True)
    measuring_period = marshmallow.fields.Integer(required=True)
    archive_period = marshmallow.fields.Integer(required=True)
    base_pressure = marshmallow.fields.Float(required=True, as_string=True)
    base_temperature = marshmallow.fields.Float(required=True, as_string=True)
    compressibility_formula = marshmallow.fields.Integer(required=True)
    gas_composition = marshmallow.fields.Nested(GasCompositionSchema, required=True)
    data_archive_record_length = marshmallow.fields.Integer(required=True)
    binary_archive_record_length = marshmallow.fields.Integer(required=True)
    daily_archive_record_length = marshmallow.fields.Integer(required=True)
    monthly_archive_record_length = marshmallow.fields.Integer(required=True)
    instantaneous_values_error_bit_order = marshmallow.fields.Integer(required=True)
    binary_archive_record_error_bit_order = marshmallow.fields.Integer(required=True)
    data_archive_record_error_bit_order = marshmallow.fields.Integer(required=True)
    optical_port_speed = marshmallow.fields.Integer(required=True)
    optical_port_protocol = marshmallow.fields.Integer(required=True)
    optical_port_bit_control = marshmallow.fields.Integer(required=True)
    communication_port_0_speed = marshmallow.fields.Integer(required=True)
    communication_port_0_protocol = marshmallow.fields.Integer(required=True)
    communication_port_0_bit_control = marshmallow.fields.Integer(required=True)
    communication_port_1_speed = marshmallow.fields.Integer(required=True)
    communication_port_1_protocol = marshmallow.fields.Integer(required=True)
    communication_port_1_bit_control = marshmallow.fields.Integer(required=True)
    communication_port_2_speed = marshmallow.fields.Integer(required=True)
    communication_port_2_protocol = marshmallow.fields.Integer(required=True)
    communication_port_2_bit_control = marshmallow.fields.Integer(required=True)
    gas_day_hour = marshmallow.fields.Integer(required=True)
    fixed_barometric_pressure = marshmallow.fields.Float(required=True, as_string=True)
    altitude = marshmallow.fields.Float(required=True, as_string=True)
    status_archive_record_length = marshmallow.fields.Integer(required=True)
    pressure_unit_type = marshmallow.fields.Integer(required=True)
    pressure_unit_text = marshmallow.fields.String(required=True)
    temperature_unit_type = marshmallow.fields.Integer(required=True)
    temperature_unit_text = marshmallow.fields.String(required=True)
    altitude_unit_type = marshmallow.fields.Integer(required=True)
    altitude_unit_text = marshmallow.fields.String(required=True)
    gross_calorific_value_type = marshmallow.fields.Integer(required=True)
    gross_calorific_value_text = marshmallow.fields.String(required=True)
    dst_region = marshmallow.fields.Integer(required=True)
    gmt_hour_shift = marshmallow.fields.Integer(required=True)
    billing_archive_record_length = marshmallow.fields.Integer(required=True)
    conditions_for_combustion_heat = marshmallow.fields.Integer(required=True)
    device_variant = marshmallow.fields.Integer(required=True)
    bit_control = marshmallow.fields.Integer(required=True)
    places_for_corrected_volume_counters = marshmallow.fields.Integer(required=True)
    device_features_bits = marshmallow.fields.Integer(required=True)
    device_features = marshmallow.fields.String(required=True)
    version_metrological_part = marshmallow.fields.String(required=True)
    metrological_crc = marshmallow.fields.Integer(required=True)
    metrological_crc_32 = marshmallow.fields.Integer(required=True)
    application_crc_32 = marshmallow.fields.Integer(required=True)

    @post_load
    def make_object(self, data, **kwargs):
        return SystemParameters(**data)



import random
from enum import IntEnum

import attr
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from elgas import utils


class EncryptionKeyId(IntEnum):
    FACTORY = 1
    TEMPORARY = 2
    ADMINISTRATOR = 3
    MAINTENANCE = 4
    USER_1 = 5
    USER_2 = 6
    USER_3 = 7


def data_padding(data: bytes):
    """
    Data needs to be padded so that data + crc is a multiple of 16 bits.
    So multiple of 2 bytes -> an even number of bytes.
    CRC is always 2 bytes so does not contribute to the evenness.

    Theis: there is a mistake in the documentation and it should be 16 bytes of
    padding and not 16 bits. this would actually give you a proper length data to
    do AES-ECB and that is why is says "no-padding".
    """
    rest = (len(data) + 2) % 16  # +2 is crc
    if not rest:
        return b""
    else:
        to_pad = 16 - rest
        padding = bytearray()
        for _ in range(0, to_pad):
            padding.append(random.randint(0, 125))

        return bytes(padding)


@attr.s(auto_attribs=True)
class CipherContext:
    key_id: EncryptionKeyId
    key: bytes

    def crc(self, data: bytes):
        for_crc = bytearray()
        for_crc.extend(len(data).to_bytes(2, "little"))  # Lenght of original data!
        for_crc.append(self.key_id)
        for_crc.extend(data)
        for_crc.extend(data_padding(data))
        return utils.calculate_crc(for_crc)

    def encrypt(self, data: bytes):
        for_crc = bytearray()
        for_crc.extend(len(data).to_bytes(2, "little"))
        for_crc.append(self.key_id)
        padded_data = data_padding(data)
        for_crc.extend(data)
        for_crc.extend(padded_data)
        crc = utils.calculate_crc(for_crc)
        unencrypted = for_crc + crc.to_bytes(2, "big")
        # we then encrypt only the data part + crc.
        header = unencrypted[:3]
        to_cipher = unencrypted[3:]

        cipher_text = encrypt(to_cipher, self.key, self.key)
        return header + cipher_text

    def decrypt(self, in_data: bytes):
        header = in_data[:3]
        encrypted = in_data[3:]
        original_length = int.from_bytes(header[:2], "little")
        key_id = header[2]
        assert key_id == self.key_id
        decrypted_data = decrypt(encrypted, self.key, self.key)
        crc = decrypted_data[-2:]
        for_crc = bytearray()
        for_crc.extend(original_length.to_bytes(2, "little"))
        for_crc.append(key_id)
        for_crc.extend(decrypted_data[:-2])
        crc = utils.calculate_crc(for_crc).to_bytes(2, "big")
        correct_crc = decrypted_data[-2:]
        if not crc == correct_crc:
            raise ValueError(
                f"Incorrect CRC in encrypted data. Got {crc} should be {correct_crc} "
            )
        original_data = decrypted_data[:original_length]
        return original_data


def encrypt(data: bytes, key: bytes, iv: bytes) -> bytes:
    encryptor = Cipher(
        algorithms.AES(key), modes.CBC(initialization_vector=iv)
    ).encryptor()
    cipher_text = encryptor.update(data) + encryptor.finalize()
    return cipher_text


def decrypt(cipher_text: bytes, key: bytes, iv: bytes):
    decryptor = Cipher(
        algorithms.AES(key), modes.CBC(initialization_vector=iv)
    ).decryptor()
    decrypted_text = decryptor.update(cipher_text) + decryptor.finalize()
    return decrypted_text



class ElgasError(Exception):
    """General Elgas error"""


class LocalElgasProtocolError(ElgasError):
    """Protocol error"""


class CommunicationError(ElgasError):
    """Communication error"""


class WrongPasswordError(ElgasError):
    # Bit 0
    """Wrong password for action was used or password switch is not turned on"""


class SettingArchiveFull(ElgasError):
    # Bit 1
    """Archive of settings is full"""


class SwitchOff(ElgasError):
    # Bit 2
    """Switch is off"""


class BlockedBuffer(ElgasError):
    # Bit 3
    """Blocked buffer"""


class DataError(ElgasError):
    # Bit 4
    """Data error"""


class CipherKeyError(ElgasError):
    # Bit 5
    """Wrong key for encryption of decryption"""


class WrongEncryptionKeysError(ElgasError):
    # Bit 6
    """Wrong encryption keys"""


class WriteError(ElgasError):
    # Bit 7
    """Write Error"""



from typing import ClassVar, Optional

import attr
import marshmallow
from marshmallow import post_load

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class Compressibility:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.COMPRESSIBILITY
    value_length: ClassVar[int] = 4

    number: int
    id: int
    address_in_actual_values: int
    address_in_data_archive_record: int
    bit_control: int
    in_data_archive: bool
    in_daily_archive: bool
    in_monthly_archive: bool
    is_metrological_quantity: bool
    name: str
    number_of_conversion_coefficient: int
    address_in_daily_archive_record: int
    address_in_monthly_archive_record: int
    decimals: Optional[int]

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = int.from_bytes(pop_many(data, 2), "little")
        id = int.from_bytes(pop_many(data, 2), "little")
        address_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        address_in_data_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_control = data.pop(0)
        in_data_archive = bool(bit_control & 0b00000001)
        in_daily_archive = bool(bit_control & 0b00000010)
        in_monthly_archive = bool(bit_control & 0b00000100)
        is_metrological_quantity = bool(bit_control & 0b00010000)
        name = pretty_text(pop_many(data, 23))
        number_of_conversion_coefficient = data.pop(0)

        address_in_daily_archive_record = int.from_bytes(pop_many(data, 2), "little")
        address_in_monthly_archive_record = int.from_bytes(pop_many(data, 2), "little")

        if data:
            decimals = data.pop(0)
        else:
            decimals = None

        return cls(
            number=number,
            id=id,
            address_in_actual_values=address_in_actual_values,
            address_in_data_archive_record=address_in_data_archive_record,
            bit_control=bit_control,
            in_data_archive=in_data_archive,
            in_daily_archive=in_daily_archive,
            in_monthly_archive=in_monthly_archive,
            is_metrological_quantity=is_metrological_quantity,
            name=name,
            number_of_conversion_coefficient=number_of_conversion_coefficient,
            address_in_daily_archive_record=address_in_daily_archive_record,
            address_in_monthly_archive_record=address_in_monthly_archive_record,
            decimals=decimals,
        )


class CompressibilityZ(Compressibility):
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.COMPRESSIBILITY_Z
    value_length: ClassVar[int] = 4


class CompressibilityZBase(Compressibility):
    object_type: ClassVar[
        ParameterObjectType
    ] = ParameterObjectType.COMPRESSIBILITY_Z_BASE
    value_length: ClassVar[int] = 4


class CompressibilitySchema(marshmallow.Schema):

    number = marshmallow.fields.Integer(required=True)
    id = marshmallow.fields.Integer(required=True)
    address_in_actual_values = marshmallow.fields.Integer(required=True)
    address_in_data_archive_record = marshmallow.fields.Integer(required=True)
    bit_control = marshmallow.fields.Integer(required=True)
    in_data_archive = marshmallow.fields.Boolean(required=True)
    in_daily_archive = marshmallow.fields.Boolean(required=True)
    in_monthly_archive = marshmallow.fields.Boolean(required=True)
    is_metrological_quantity = marshmallow.fields.Boolean(required=True)
    name = marshmallow.fields.String(required=True)
    number_of_conversion_coefficient = marshmallow.fields.Integer(required=True)
    address_in_daily_archive_record = marshmallow.fields.Integer(required=True)
    address_in_monthly_archive_record = marshmallow.fields.Integer(required=True)
    decimals = marshmallow.fields.Integer(required=True, allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return Compressibility(**data)


class CompressibilityZSchema(CompressibilitySchema):
    @post_load
    def make_object(self, data, **kwargs):
        return CompressibilityZ(**data)


class CompressibilityZBaseSchema(CompressibilitySchema):
    @post_load
    def make_object(self, data, **kwargs):
        return CompressibilityZBase(**data)



from typing import ClassVar

import attr
import marshmallow
from marshmallow import post_load

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class Diagnostics:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.DIAGNOSTICS

    number: int
    id: int
    address_in_actual_values: int
    address_in_data_archive_record: int
    bit_control: int
    in_data_archive: bool
    in_daily_archive: bool
    in_monthly_archive: bool
    in_factory_archive: bool
    name: str

    address_in_daily_archive_record: int
    address_in_monthly_archive_record: int
    mask_1_of_status_archive: int
    mask_2_of_status_archive: int
    mask_1_of_alarm: int
    mask_2_of_alarm: int
    mask_1_of_calling_to_dispatching: int
    mask_2_of_calling_to_dispatching: int
    action_during_change: int

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = int.from_bytes(pop_many(data, 2), "little")
        id = int.from_bytes(pop_many(data, 2), "little")
        address_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        address_in_data_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_control = data.pop(0)
        in_data_archive = bool(bit_control & 0b00000001)
        in_daily_archive = bool(bit_control & 0b00000010)
        in_monthly_archive = bool(bit_control & 0b00000100)
        in_factory_archive = bool(bit_control & 0b00001000)
        name = pretty_text(pop_many(data, 23))

        address_in_daily_archive_record = int.from_bytes(pop_many(data, 2), "little")
        address_in_monthly_archive_record = int.from_bytes(pop_many(data, 2), "little")
        mask_1_of_status_archive = int.from_bytes(pop_many(data, 4), "little")
        mask_2_of_status_archive = int.from_bytes(pop_many(data, 4), "little")
        mask_1_of_alarm = int.from_bytes(pop_many(data, 4), "little")
        mask_2_of_alarm = int.from_bytes(pop_many(data, 4), "little")
        mask_1_of_calling_to_dispatching = int.from_bytes(pop_many(data, 4), "little")
        mask_2_of_calling_to_dispatching = int.from_bytes(pop_many(data, 4), "little")
        action_during_change = data.pop(0)

        return cls(
            number=number,
            id=id,
            address_in_actual_values=address_in_actual_values,
            address_in_data_archive_record=address_in_data_archive_record,
            bit_control=bit_control,
            in_data_archive=in_data_archive,
            in_daily_archive=in_daily_archive,
            in_monthly_archive=in_monthly_archive,
            in_factory_archive=in_factory_archive,
            name=name,
            address_in_daily_archive_record=address_in_daily_archive_record,
            address_in_monthly_archive_record=address_in_monthly_archive_record,
            mask_1_of_status_archive=mask_1_of_status_archive,
            mask_2_of_status_archive=mask_2_of_status_archive,
            mask_1_of_alarm=mask_1_of_alarm,
            mask_2_of_alarm=mask_2_of_alarm,
            mask_1_of_calling_to_dispatching=mask_1_of_calling_to_dispatching,
            mask_2_of_calling_to_dispatching=mask_2_of_calling_to_dispatching,
            action_during_change=action_during_change,
        )


class DiagnosticsSchema(marshmallow.Schema):

    number = marshmallow.fields.Integer(required=True)
    id = marshmallow.fields.Integer(required=True)
    address_in_actual_values = marshmallow.fields.Integer(required=True)
    address_in_data_archive_record = marshmallow.fields.Integer(required=True)
    bit_control = marshmallow.fields.Integer(required=True)
    in_data_archive = marshmallow.fields.Boolean(required=True)
    in_daily_archive = marshmallow.fields.Boolean(required=True)
    in_monthly_archive = marshmallow.fields.Boolean(required=True)
    in_factory_archive = marshmallow.fields.Boolean(required=True)
    name = marshmallow.fields.String(required=True)

    address_in_daily_archive_record = marshmallow.fields.Integer(required=True)
    address_in_monthly_archive_record = marshmallow.fields.Integer(required=True)
    mask_1_of_status_archive = marshmallow.fields.Integer(required=True)
    mask_2_of_status_archive = marshmallow.fields.Integer(required=True)
    mask_1_of_alarm = marshmallow.fields.Integer(required=True)
    mask_2_of_alarm = marshmallow.fields.Integer(required=True)
    mask_1_of_calling_to_dispatching = marshmallow.fields.Integer(required=True)
    mask_2_of_calling_to_dispatching = marshmallow.fields.Integer(required=True)
    action_during_change = marshmallow.fields.Integer(required=True)

    @post_load
    def make_object(self, data, **kwargs):
        return Diagnostics(**data)
